@@ -1,27 +1,26 @@
-import decimal
+from typing import TypedDict, Callable, TypeVar
 import sys
-from collections import namedtuple
-from typing import TypedDict, Callable, TypeVar, Sequence
+from enum import Enum
+import decimal
 from datetime import date, datetime
 import asyncio
 
 from graphql import GraphQLError
 from graphql.type.definition import GraphQLObjectType, GraphQLEnumType, get_named_type, is_list_type, is_non_null_type
 from ariadne.types import GraphQLResolveInfo
-from sqlalchemy import Row, ColumnElement, select, Executable, cast, String, Select, and_
-from sqlalchemy.orm import DeclarativeBase, aliased
-from sqlalchemy.ext.asyncio import AsyncConnection
-from sqlalchemy.exc import MultipleResultsFound
+import pypika
+from pypika import functions as func
+from pypika.queries import QueryBuilder
 from pyparsing import ParseException
-from aiodataloader import DataLoader
 
-from .auth_utils import User
-from .models_utils import Relation
+from .request_utils import get_db
+from .models_utils import BaseModel, get_field_main_type
 from .graphql_app import check_permission, HideResult
 from .filter_parser import number_filter_parser, date_filter_parser, datetime_filter_parser, boolean_filter_parser
+from . import pypika_extentions as func2
 
-M = TypeVar('M', bound=DeclarativeBase)
-nulldate = date(1753, 1, 1)  # For Incadea Tables
+
+M = TypeVar('M')
 
 
 class FieldFilter(TypedDict):
@@ -29,65 +28,13 @@ class FieldFilter(TypedDict):
     value: str
 
 
-# Shortcuts
+class PyodbcParams:
+    def __init__(self):
+        self.params = []
 
-def get_table_primary_keys(model: type[M]) -> list[ColumnElement]:
-    return [col for col in model.__mapper__.columns if col.primary_key]
-
-
-def filter_primary_key(q: Executable, model: type[M], primary_key_values: list[str]):
-    pk_cols = get_table_primary_keys(model)
-    try:
-        for i, pk_val in enumerate(primary_key_values):
-            q = q.where(pk_cols[i] == pk_val)
-    except IndexError:
-        raise GraphQLError(f"You have specified more primary keys values than there are columns for model {model.__name__}")
-    return q
-
-
-def add_typename_to_rows(model: type[DeclarativeBase], rows: Sequence[Row], type_name=None) -> list[namedtuple]:
-    if type_name is None:
-        type_name = model.__name__
-    keys = [col.key or col.name for col in model.__mapper__.columns]
-    NamedTuple = namedtuple(type_name, keys)
-    NamedTuple.__typename = type_name
-    return [NamedTuple(*(getattr(row, key) for key in keys)) for row in rows]
-
-
-def get_user(info: GraphQLResolveInfo) -> User:
-    return info.context['request'].user
-
-
-def get_db(info) -> AsyncConnection:
-    return info.context['request'].state.db
-
-
-def get_dataloader(info: GraphQLResolveInfo, loader_key: str, batch_function_wrapper):
-    state = info.context['request'].state
-    if not hasattr(state, 'loaders'):
-        state.loaders = {}
-    return state.loaders.setdefault(loader_key, DataLoader(batch_function_wrapper(get_db(info))))
-
-
-async def get_or_none(db: AsyncConnection, selectable: type[M] | ColumnElement, primary_keys: list):
-    column_mode = getattr(selectable, 'is_attribute', False)  # Guess if Column
-    model = selectable.class_ if column_mode else selectable
-    result = await db.execute(filter_primary_key(select(selectable), model, primary_keys))
-    if column_mode:
-        return result.scalar_one_or_none()
-    else:
-        return result.one_or_none()
-
-
-async def get_or_error(db: AsyncConnection, selectable: type[M] | ColumnElement, primary_keys: list):
-    try:
-        row: M | None = await get_or_none(db, selectable, primary_keys)
-    except MultipleResultsFound:
-        raise GraphQLError(f"Primary keys: {primary_keys} returned more than one row")
-    if row is None:
-        model = getattr(selectable, 'class_', selectable)
-        raise GraphQLError(f"Could not find {model.__name__}: {primary_keys}")
-    return row
+    def add(self, value):
+        self.params.append(value)
+        return pypika.Parameter('?')
 
 
 # Resolver Tools
@@ -104,20 +51,11 @@ def separate_filters(filters: list[FieldFilter], field_names_to_separate: list[s
     return newfilters, separated
 
 
-def model_to_dict(model: DeclarativeBase, remove_fields: list[str] = None):
-    if remove_fields is None:
-        remove_fields = []
-    return {
-        field: getattr(model, field)
-        for field in model.__mapper__.column_attrs.keys()
-        if field not in remove_fields
-    }
-
-
 # Complete Resolvers
 
-def resolve_type_inspector_factory(model_base: type[DeclarativeBase]):
-    models_dict = {model.class_.__name__: model.class_ for model in model_base.registry.mappers}
+def resolve_type_inspector_factory(model_list: list[BaseModel]):
+    models_dict = {model.__name__: model for model in model_list}
+
     # Todo: Deduce filtering automatically
     def resolve_type_inspector(_, info: GraphQLResolveInfo, type_name: str):
         gqltype = info.schema.get_type(type_name)
@@ -126,9 +64,9 @@ def resolve_type_inspector_factory(model_base: type[DeclarativeBase]):
 
         # Primary Keys. Raise error when using directive and model not found, else ignore
         model_name = getattr(gqltype, '__modelname__', type_name)
-        model: type[DeclarativeBase] | None = models_dict.get(model_name)
+        model: type[BaseModel] | None = models_dict.get(model_name)
         if model is not None:
-            primary_keys = [col.name for col in get_table_primary_keys(model)]
+            primary_keys = model.primary_keys
         elif hasattr(gqltype, '__modelname__'):
             raise GraphQLError(f"Could not find model with name {model_name}")
         else:
@@ -173,28 +111,12 @@ def resolve_type_inspector_factory(model_base: type[DeclarativeBase]):
     return resolve_type_inspector
 
 
-def apply_model_relation(q: Select, model: type[DeclarativeBase], relation_name: str) -> tuple[Select, type[DeclarativeBase]]:
-    try:
-        relation: Relation = getattr(model, '__relations__', {})[relation_name]
-    except KeyError:
-        raise GraphQLError(f"Could not find Relation {relation_name} in model {model.__name__}")
+def load_from_model_query(
+        model: type[BaseModel], filters: list[FieldFilter], limit: int | None, offset: int | None,
+        query_modifier: Callable[[QueryBuilder], QueryBuilder] | None = None, init_query: QueryBuilder | None = None
+) -> QueryBuilder:
 
-    rel_model = aliased(relation.model)
-
-    if relation.join_on is None:
-        q = q.join(rel_model)
-    else:
-        q = q.join(rel_model, and_(*(
-            getattr(rel_model, foreign_col) == local_col
-            for foreign_col, local_col in relation.join_on.items())
-        ))
-
-    return q, rel_model
-
-
-def load_from_model_query(model: type[M], filters: list[FieldFilter], limit: int | None, offset: int | None,
-                          query_modifier: Callable[[Executable], Executable] | None = None, init_query: Executable | None = None) -> Executable:
-    q = select(model) if init_query is None else init_query
+    q = model.registry.query_class.from_(model.t).select(*model.fs) if init_query is None else init_query
 
     for f in filters:
         full_name = f['field_name']
@@ -204,15 +126,15 @@ def load_from_model_query(model: type[M], filters: list[FieldFilter], limit: int
         current_model = model
         for relation_name in relation_names:
             # Get Relation model and join
-            q, current_model = apply_model_relation(q, current_model, relation_name)
+            q, current_model = current_model.join_relation(q, relation_name)
 
-        field = getattr(current_model, field_name)
-        field_type = field.type.python_type
+        field: pypika.Field = getattr(current_model.f, field_name)
+        field_type = get_field_main_type(current_model, field_name)
 
         # Deducing filter type by model column type. Contrary to resolve_type_inspector.
         try:
-            if field_type is str:
-                q = q.where(cast(field, String).ilike(value))  # cast used to make Enum behave like strings.
+            if field_type is str or issubclass(field_type, Enum):
+                q = q.where(func.Lower(func.Cast(field, 'varchar')).like(value.lower()))  # cast used to make Enum behave like strings.
             elif field_type in [int, float, decimal.Decimal]:
                 q = number_filter_parser(q, field, value)
             elif field_type is date:
@@ -222,11 +144,11 @@ def load_from_model_query(model: type[M], filters: list[FieldFilter], limit: int
             elif field_type is bool:
                 q = boolean_filter_parser(q, field, value)
             elif field_type is list:
-                q = q.where(field.any_() == value)
+                q = q.where(func2.Any(value, field))
             else:
-                raise GraphQLError(f'Cannot filter on column type {field_type}')
+                raise GraphQLError(f"Cannot filter on column type {field_type}")
         except ParseException as e:
-            raise GraphQLError(f'Cannot parse value: {value} for field {field} of type {field_type} [{e}]')
+            raise GraphQLError(f"Cannot parse value: {value} for field {field} of type {field_type} [{e}]")
 
     if query_modifier is not None:
         q = query_modifier(q)
@@ -234,10 +156,10 @@ def load_from_model_query(model: type[M], filters: list[FieldFilter], limit: int
     return q.limit(limit).offset(offset)
 
 
-def resolve_type_loader_factory(model_base: type[DeclarativeBase]):
-    models_dict = {model.class_.__name__: model.class_ for model in model_base.registry.mappers}
+def resolve_type_loader_factory(model_list: list[BaseModel]):
+    models_dict = {model.__name__: model for model in model_list}
 
-    async def resolve_type_loader(_, info, type_name, filters: list[FieldFilter], limit: int, offset: int):
+    async def resolve_type_loader(_, info, type_name: str, filters: list[FieldFilter], limit: int, offset: int):
         gqltype = info.schema.get_type(type_name)
         if gqltype is None:  # Check if Type exists in GQL
             raise GraphQLError(f'Type {type_name} does not exist')
@@ -248,17 +170,33 @@ def resolve_type_loader_factory(model_base: type[DeclarativeBase]):
         except KeyError:
             raise GraphQLError(f"Could not find {model_name} in Models")
 
-        q = load_from_model_query(model, filters, limit, offset)
-        rows = (await get_db(info).execute(q)).all()
-        return add_typename_to_rows(model, rows, type_name)
+        registry = model.registry
+
+        q = load_from_model_query(model, filters, limit, offset).orderby(
+            *(model.database_fields[pkname] for pkname in model.primary_keys)
+        )
+
+        if registry.custom_loader is None:
+            recs = await get_db(info).fetch(q)
+        else:
+            recs = await registry.custom_loader(info, q)
+
+        objs = model.build_all(recs)
+        for obj in objs:
+            obj.__typename = type_name
+        return objs
 
     return resolve_type_loader
 
 
-def simple_table_resolver_factory(model: type[M], query_modifiers=None):
+def simple_table_resolver_factory(model: type[M], query_modifiers: Callable[[QueryBuilder], QueryBuilder] | None = None):
     async def simple_table_resolver(_, info, filters: list[FieldFilter], limit: int, offset: int) -> list[M]:
         q = load_from_model_query(model, filters, limit, offset, query_modifiers)
-        return (await get_db(info).execute(q)).all()
+        if model.registry.custom_loader is None:
+            recs = await get_db(info).fetch(q)
+        else:
+            recs = await model.registry.custom_loader(info, q)
+        return model.build_all(recs)
     return simple_table_resolver
 
 
